#include <io.h>
#include <stdio.h>
#include <direct.h>
#include <windows.h>
#include <io.h>
#include <sys/stat.h>
#include "mpc\in_mpc.h"
#include "mpc\requant.h"
#include "mpc\huffsv46.h"
#include "mpc\huffsv7.h"
#include "mpc\bitstream.h"
#include "mpc\mpc_dec.h"
#include "mpc\minimax.h"

#ifndef _MAX_PATH
#define _MAX_PATH   260
#endif

void usage ()
{
    printf ( "Usage: MPCscan [options] [directory/filename]\n"
             "Scans for corruptions in Musepack files.\n"
             "Default operation is to scan all files in current working directory.\n"
             "\n"
             "Options:\n"
             "-r: recursive\n"
           );
}

int print_time ( char* str, unsigned int th, unsigned int tm, unsigned int ts )
{
    if ( th > 0 )
        return sprintf ( str, "%02u:%02u:%02u", th, tm, ts ); 
    else
        return sprintf ( str, "%02u:%02u", tm, ts ); 
}

void print_fn ( const char* fn, const char* path )
{
    if ( path && path[0] )
        printf ( "'%s\\%s'", path, fn );
    else
        printf ( "'%s'", fn );
}

void titlebar ( const char* fn, const char* path, unsigned int progress )
{
    char temp[512];

    if ( path && path[0] )
        sprintf ( temp, "%02u%% - %s\\%s", progress, path, fn );
    else
        sprintf ( temp, "%02u%% - %s", progress, fn );

    SetConsoleTitle ( temp );
}

static unsigned __int64 bytes = 0;

unsigned int scan_mpc ( const char* fn, const char* path, unsigned int* files, unsigned int* errors, unsigned __int64 totalsize )
{
    MPC_decoder *theDec = NULL;
    Reader* reader = NULL;
    StreamInfo *Info = NULL;
    bool broken = FALSE;
    unsigned int progress;

    if ( totalsize > 0 ) {
        unsigned __int64 t = bytes * 100 / totalsize;
        progress = (unsigned int)t;
    } else {
        progress = 100;
    }

    *files += 1;

    titlebar ( fn, path, progress );

    FILE *fp = fopen ( fn, "rb" );
    if ( !fp ) {
        print_fn ( fn, path );
        printf ( ": file open failed\n" );
        *errors += 1;
        return 1;
    }

    reader = new Reader_file ( fp );
    Info = new StreamInfo ();
    if ( Info->ReadStreamInfo (reader) != 0 ) {
        print_fn ( fn, path );
        printf ( ": unrecognized format\n" );
        if ( reader ) delete reader;
        if ( Info ) delete Info;
        *errors += 1;
        return 1;
    }

    theDec = new MPC_decoder ( reader );
    theDec->RESET_Globals   ();
    theDec->RESET_Synthesis ();
    theDec->SetStreamInfo ( Info );
    theDec->FileInit ();

    if ( Info->simple.HeaderPosition > 0 ) {
        print_fn ( fn, path );
        printf ( ": invalid MPC, contains ID3v2 tag!\n" );
    }

    unsigned int lastpos = -1;
    unsigned int pos;

    while ( 1 ) {
        unsigned int RING = theDec->Zaehler;
        int valid_samples = theDec->DECODE ( NULL/*sample_buffer*/ );
        theDec->UpdateBuffer ( RING );
        if ( theDec->FrameWasValid == 0 ) {
            broken = TRUE;
            break;
        }
        if ( valid_samples <= 0 ) break;

        if ( totalsize > 0 ) {
            unsigned __int64 t = (bytes + (theDec->BitsRead()/8)) * 100 / totalsize;
            pos = (unsigned int)t;

            if ( pos != lastpos ) {
                lastpos = pos;
                titlebar ( fn, path, pos );
            }
        }
    }

    bytes += Info->simple.TotalFileLength;

    if ( broken ) {
        char time1[32];
        unsigned long ts = (unsigned long)(theDec->DecodedFrames * FRAMELEN / Info->simple.SampleFreq);
        unsigned long tm = (ts/60) % 60;
        unsigned long th = (ts/60/60);
        ts = ts % 60;
        print_time ( time1, th, tm, ts );

        print_fn ( fn, path );
        printf ( ": broken after frame %u/%u (%s)\n", theDec->DecodedFrames, Info->simple.Frames, time1 );
        *errors += 1;
    }

    if ( Info ) delete Info;
    if ( reader ) delete reader;
    if ( theDec ) delete theDec;


    return broken;
}

unsigned int get_size ( const char* fn, unsigned int* files, unsigned __int64* totalsize )
{
    *files += 1;

    FILE *fp = fopen ( fn, "rb" );
    if ( !fp ) return 0;
    fseek ( fp, 0, SEEK_END );
    long size = ftell ( fp );
    fclose ( fp );

    *totalsize += size;

    return size;
}

unsigned int scan_files ( unsigned int *files, unsigned int *errors, unsigned __int64 *totalsize, const char* path, BOOL onlycheck )
{
    struct _finddata_t  fd;
    long                hFile;

    if ( (hFile = _findfirst ("*.*", &fd)) != -1L ) {
        if ( !(fd.attrib & _A_SUBDIR) && !(fd.attrib & _A_HIDDEN) ) {
            char *p = strrchr ( fd.name, '.' );
            if ( p ) {
                if ( !stricmp (p, ".mpc") || !stricmp (p, ".mp+") || !stricmp (p, ".mpp") ) {
                    if ( !onlycheck ) {
                        scan_mpc ( fd.name, path, files, errors, *totalsize );
                    } else {
                        get_size ( fd.name, files, totalsize );
                    }
                }
            }
        }

        while ( _findnext (hFile, &fd) == 0 ) {
            if ( fd.attrib & _A_SUBDIR || fd.attrib & _A_HIDDEN )
                continue;

            char *p = strrchr ( fd.name, '.' );
            if ( p ) {
                if ( !stricmp (p, ".mpc") || !stricmp (p, ".mp+") || !stricmp (p, ".mpp") ) {
                    if ( !onlycheck ) {
                        scan_mpc ( fd.name, path, files, errors, *totalsize );
                    } else {
                        get_size ( fd.name, files, totalsize );
                    }
                }
            }
        }

        _findclose ( hFile );
    }

    return 0;
}

unsigned int scan_path ( const char* path, int recursive, unsigned int *files, unsigned int *errors, unsigned __int64 *totalsize, BOOL onlycheck )
{
    char                start_path [_MAX_PATH];
    char                temp_path [_MAX_PATH];
    struct _finddata_t  fd;
    long                hFile;

    _getcwd ( start_path, _MAX_PATH );
    if ( path != NULL ) _chdir  ( path );
    _getcwd ( temp_path, _MAX_PATH );
    if ( temp_path[0] != '\0' && temp_path[strlen (temp_path)-1] == '\\' )
        temp_path[strlen (temp_path)-1] = '\0';
    scan_files ( files, errors, totalsize, temp_path, onlycheck );

    if ( recursive ) {
        // go through all subdirectories
        if ( (hFile = _findfirst ("*.*", &fd)) != -1L ) {
            if ( !(strcmp (fd.name, ".") == 0 || strcmp (fd.name, "..") == 0) ) {
                if ( fd.attrib & _A_SUBDIR && !(fd.attrib & _A_HIDDEN) ) {
                    //printf ( "\nEntering directory: '%s'\n", fd.name );
                    scan_path ( fd.name, recursive, files, errors, totalsize, onlycheck );
                }
            }

            while ( _findnext (hFile, &fd) == 0 ) {
                if ( strcmp (fd.name, ".") == 0 || strcmp (fd.name, "..") == 0 )
                    continue;

                if ( fd.attrib & _A_SUBDIR && !(fd.attrib & _A_HIDDEN) ) {
                    //printf ( "\nEntering directory: '%s'\n", fd.name );
                    scan_path ( fd.name, recursive, files, errors, totalsize, onlycheck );
                }
            }

            _chdir ( start_path );
            _findclose ( hFile );
        }
    } else {
        _chdir ( start_path );
    }

    return 0;
}

int isdir ( const char* path )
{
    struct _stat st;
    if ( _stat (path, &st) != 0 ) return 0;
    return st.st_mode & _S_IFDIR;
}

int isfile ( const char* fn )
{
    FILE* fp = fopen ( fn, "rb" );
    if ( !fp ) return 0;
    fclose ( fp );
    return 1;
}

void setpriority ()
{
    HANDLE  pid = GetCurrentProcess ();
    HANDLE  tid = GetCurrentThread  ();

    SetPriorityClass ( pid, IDLE_PRIORITY_CLASS );
    SetThreadPriority ( tid, THREAD_PRIORITY_IDLE );
}

int main ( int argc, char** argv )
{
    int                 recursive = 0;
    int                 scanfiles = 1;
    unsigned int        files = 0;
    unsigned int        errors = 0;
    unsigned __int64    totalsize = 0;

    setpriority ();

    for ( int onlycheck = 1; onlycheck >= 0; onlycheck-- ) {
        files = 0;
        errors = 0;
        recursive = 0;
        scanfiles = 1;

        if ( argc > 1 ) {
            // show usage if first parameter is not '-r', file or directory
            if ( strcmp (argv[1], "-r") && !(isdir (argv[1]) || isfile (argv[1])) ) {
                usage ();
                return 0;
            }

            for ( int i = 1; i < argc; i++ ) {
                if ( !strcmp (argv[i], "-r") ) {    // '-r' option (recursive)
                    recursive = 1;
                } else
                if ( isdir (argv[i]) ) {            // directory
                    scan_path ( argv[i], recursive, &files, &errors, &totalsize, onlycheck );
                    scanfiles = 0;
                } else
                if ( isfile (argv[i]) ) {           // file
                    if ( !onlycheck ) {
                        scan_mpc ( argv[i], NULL, &files, &errors, totalsize );
                    } else {
                        get_size ( argv[i], &files, &totalsize );
                    }
                    scanfiles = 0;
                }
            }
        }

        if ( scanfiles ) {
            scan_path ( NULL, recursive, &files, &errors, &totalsize, onlycheck );
        }
    }

    printf ( "\nScanned %u file(s)\nErrors with %u\n", files, errors );

    return errors != 0;
}
